"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getReferenceCountCodeLenses = getReferenceCountCodeLenses;
const references_1 = require("./references");
const translation_1 = require("./utils/translation");
function getReferenceCountCodeLenses(analyzer, document) {
    const codeLenses = [];
    // Filter for global symbols
    const globalSymbols = analyzer.getFlatDocumentSymbols(document.uri)
        .filter(symbol => symbol.fishKind === 'FUNCTION');
    // Create a code lens for each global symbol
    for (const symbol of globalSymbols) {
        // Get reference count
        const references = (0, references_1.getReferences)(analyzer, document, symbol.selectionRange.start) || [];
        const referencesCount = references.length;
        codeLenses.push({
            range: symbol.range,
            command: {
                title: `${referencesCount} references`,
                command: 'fish-lsp.showReferences',
                arguments: [(0, translation_1.uriToPath)(document.uri), symbol.selectionRange.start, references],
            },
        });
        // Create code lens with count display
        // codeLenses.push({
        //   range: symbol.selectionRange,
        //   command: {
        //     title: `${referencesCount} references`,
        //     command: "fish-lsp.showReferences",
        //     arguments: [document.getFilePath(), symbol.selectionRange.start, references]
        //   }
        // });
    }
    return codeLenses;
}
