"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ArparseOptions = void 0;
exports.findArgparseOptions = findArgparseOptions;
exports.findArgparseDefinitionNames = findArgparseDefinitionNames;
exports.isArgparseVariableDefinitionName = isArgparseVariableDefinitionName;
exports.convertNodeRangeWithPrecedingFlag = convertNodeRangeWithPrecedingFlag;
exports.isGlobalArgparseDefinition = isGlobalArgparseDefinition;
exports.getGlobalArgparseLocations = getGlobalArgparseLocations;
exports.getArgparseDefinitionName = getArgparseDefinitionName;
exports.isCompletionArgparseFlagWithCommandName = isCompletionArgparseFlagWithCommandName;
exports.processArgparseCommand = processArgparseCommand;
const node_types_1 = require("../utils/node-types");
const options_1 = require("./options");
const symbol_1 = require("./symbol");
const definition_scope_1 = require("../utils/definition-scope");
const tree_sitter_1 = require("../utils/tree-sitter");
const path_1 = __importStar(require("path"));
const file_operations_1 = require("../utils/file-operations");
const translation_1 = require("../utils/translation");
const workspace_manager_1 = require("../utils/workspace-manager");
exports.ArparseOptions = [
    options_1.Option.create('-n', '--name').withValue(),
    options_1.Option.create('-x', '--exclusive').withValue(),
    options_1.Option.create('-N', '--min-args').withValue(),
    options_1.Option.create('-X', '--max-args').withValue(),
    options_1.Option.create('-i', '--ignore-unknown'),
    options_1.Option.create('-s', '--stop-nonopt'),
    options_1.Option.create('-h', '--help'),
];
const isBefore = (a, b) => a.startIndex < b.startIndex;
function findArgparseOptions(node) {
    if ((0, node_types_1.isCommandWithName)(node, 'argparse'))
        return undefined;
    const endChar = node.children.find(node => (0, node_types_1.isEndStdinCharacter)(node));
    if (!endChar)
        return undefined;
    const nodes = node.childrenForFieldName('argument')
        .filter(n => !(0, node_types_1.isEscapeSequence)(n) && isBefore(n, endChar))
        .filter(n => !(0, node_types_1.isVariableExpansion)(n) || n.type !== 'variable_name');
    return (0, options_1.findOptions)(nodes, exports.ArparseOptions);
}
function findArgparseDefinitionNames(node) {
    // check if the node is a 'argparse' command
    if (!node || !(0, node_types_1.isCommandWithName)(node, 'argparse'))
        return [];
    // check if the node has a '--' token
    const endChar = node.children.find(node => (0, node_types_1.isEndStdinCharacter)(node));
    if (!endChar)
        return [];
    // get the children of the node that are not options and before the endChar (currently skips variables)
    const nodes = node.childrenForFieldName('argument')
        .filter(n => !(0, node_types_1.isEscapeSequence)(n) && isBefore(n, endChar))
        .filter(n => !(0, node_types_1.isVariableExpansion)(n) || n.type !== 'variable_name');
    const { remaining } = (0, options_1.findOptions)(nodes, exports.ArparseOptions);
    return remaining;
}
/**
 * Checks if a node is an `argparse` definition variable
 * NOTE: if the node in question is a variable expansion, it will be skipped.
 * ```fish
 * argparse --max-args=2 --ignore-unknown --stop-nonopt h/help 'n/name=?' 'x/exclusive' -- $argv
 * ```
 * Would return true for the following SyntaxNodes passed in:
 * - `h/help`
 * - `n/name=?`
 * - `x/exclusive`
 * @param node The node to check, where it's parent isCommandWithName(parent, 'argparse'), and it's not a switch
 * @returns true if the node is an argparse definition variable (flags for the `argparse` command with be skipped)
 */
function isArgparseVariableDefinitionName(node) {
    if (!node.parent || !(0, node_types_1.isCommandWithName)(node.parent, 'argparse'))
        return false;
    const children = findArgparseDefinitionNames(node.parent);
    return !!children.some(n => n.equals(node));
}
function convertNodeRangeWithPrecedingFlag(node) {
    const range = (0, tree_sitter_1.getRange)(node);
    if (node.text.startsWith('_flag_')) {
        range.start = {
            line: range.start.line,
            character: range.start.character + 6,
        };
    }
    return range;
}
function isGlobalArgparseDefinition(analyzer, document, symbol) {
    if (symbol.fishKind !== 'ARGPARSE')
        return false;
    const parent = symbol.scopeNode;
    if (parent.type === 'function_definition') {
        const functionName = parent.firstNamedChild?.text;
        if (document.getAutoLoadName() !== functionName) {
            return false;
        }
        const filepath = (0, translation_1.uriToPath)(document.uri);
        // const workspaceDirectory = workspaces.find(ws => ws.contains(filepath) || ws.path === filepath)?.path || dirname(dirname(filepath));
        const workspaceDirectory = workspace_manager_1.workspaceManager.findContainingWorkspace(document.uri)?.path || (0, path_1.dirname)((0, path_1.dirname)(filepath));
        const completionFile = document.getAutoloadType() === 'conf.d' || document.getAutoloadType() === 'config'
            ? document.getFilePath()
            : path_1.default.join(workspaceDirectory, 'completions', document.getFilename());
        if (process.env.NODE_ENV !== 'test' && !file_operations_1.SyncFileHelper.isFile(completionFile)) {
            return false;
        }
        return analyzer.getFlatCompletionSymbols((0, translation_1.pathToUri)(completionFile)).length > 0;
    }
    return false;
}
/**
 * This is really more of a utility to ensure that any document that would contain
 * any references to completions for an autoloaded file, is parsed by the analyzer.
 */
function getGlobalArgparseLocations(analyzer, document, symbol) {
    if (isGlobalArgparseDefinition(analyzer, document, symbol)) {
        const filepath = (0, translation_1.uriToPath)(document.uri);
        const workspaceDirectory = workspace_manager_1.workspaceManager.findContainingWorkspace(document.uri)?.path || (0, path_1.dirname)((0, path_1.dirname)(filepath));
        const completionFile = document.getAutoloadType() === 'conf.d' || document.getAutoloadType() === 'config'
            ? document.getFilePath()
            : path_1.default.join(workspaceDirectory, 'completions', document.getFilename());
        if (process.env.NODE_ENV !== 'test' && !file_operations_1.SyncFileHelper.isFile(completionFile)) {
            return [];
        }
        return analyzer
            .getFlatCompletionSymbols((0, translation_1.pathToUri)(completionFile))
            .filter(s => s.equalsArgparse(symbol))
            .map(s => s.toLocation());
    }
    return [];
}
function getArgparseScopeModifier(document, _node) {
    const autoloadType = document.getAutoloadType();
    switch (autoloadType) {
        case 'conf.d':
        case 'config':
        case 'functions':
            return 'local';
        default:
            // return isTopLevelDefinition(node) ? 'global' : 'local';
            return 'local';
    }
}
function getArgparseDefinitionName(node) {
    if (!node.parent || !(0, node_types_1.isCommandWithName)(node.parent, 'complete'))
        return '';
    if (node.text) {
        const text = `_flag_${node.text}`;
        return text.replace(/-/, '_');
    }
    return '';
}
/**
 * Checks if a syntax node is a completion argparse flag with a specific command name.
 *
 * On the input: `complete -c test -s h -l help -d 'show help info for the test command'`
 *                                          ^---- node is here
 * A truthy result would be returned from the following function call:
 *
 * `isCompletionArgparseFlagWithCommandName(node, 'test', 'help')`
 * ___
 * @param node - The syntax node to check
 * @param commandName - The command name to match against
 * @param flagName - The flag name to match against
 * @param opts - Optional configuration options
 * @param opts.noCommandNameAllowed - When true, a completion without a `-c`/`--command` Option is allowed
 * @param opts.discardIfContainsOptions - A list of options that, if present, will cause the match to be discarded
 * @returns True if the node is a completion argparse flag with the specified command name
 */
function isCompletionArgparseFlagWithCommandName(node, commandName, flagName, opts) {
    // make sure that the node we are checking is inside a completion definition
    if (!node?.parent || !(0, node_types_1.isCommandWithName)(node.parent, 'complete'))
        return false;
    // parent is the entire completion command
    const parent = node.parent;
    // check if any of the options to discard are seen
    if (opts?.discardIfContainsOptions) {
        for (const option of opts.discardIfContainsOptions) {
            if (parent.children.some(c => option.matches(c))) {
                return false;
            }
        }
    }
    // check if the command name is present in the completion
    let completeCmdName = !!parent.children.find(c => c.previousSibling &&
        (0, options_1.isMatchingOption)(c.previousSibling, options_1.Option.create('-c', '--command')) &&
        c.text === commandName);
    // if noCommandNameAllowed is true, and we don't have a command name yet
    // update the completeCmdName to be true if the `-c`/`--command` option is not present
    if (opts?.noCommandNameAllowed && !completeCmdName) {
        completeCmdName = !parent.children.some(c => c.previousSibling &&
            (0, options_1.isMatchingOption)(c.previousSibling, options_1.Option.create('-c', '--command')));
    }
    // Here we determine if which type of option we are looking for
    const option = flagName.length === 1
        ? options_1.Option.create('-s', '--short')
        : options_1.Option.create('-l', '--long');
    // check if the option name is present in the completion
    const completeFlagName = !!(node.previousSibling &&
        option.equals(node.previousSibling) &&
        node.text === flagName);
    // return true if both the command name and option name
    return completeCmdName && completeFlagName;
}
function createSelectionRange(node, flags, flag, idx) {
    const range = (0, tree_sitter_1.getRange)(node);
    const text = node.text;
    const shortenedFlag = flag.replace(/^_flag_/, '');
    if (flags.length === 2 && idx === 0) {
        if ((0, node_types_1.isString)(node)) {
            range.start = {
                line: range.start.line,
                character: range.start.character + 1,
            };
            range.end = {
                line: range.start.line,
                character: range.start.character - 1,
            };
        }
        return {
            start: range.start,
            end: {
                line: range.start.line,
                character: range.start.character + shortenedFlag.length,
            },
        };
    }
    else if (flags.length === 2 && idx === 1) {
        return {
            start: {
                line: range.start.line,
                character: range.start.character + text.indexOf('/') + 1,
            },
            end: {
                line: range.end.line,
                character: range.start.character + text.indexOf('/') + 1 + shortenedFlag.length,
            },
        };
    }
    else if (flags.length === 1) {
        if ((0, node_types_1.isString)(node)) {
            return {
                start: {
                    line: range.start.line,
                    character: range.start.character + 1,
                },
                end: {
                    line: range.start.line,
                    character: range.start.character + 1 + shortenedFlag.length,
                },
            };
        }
        else {
            return (0, tree_sitter_1.getRange)(node);
        }
    }
    return range;
}
// split the `h/help` into `h` and `help`
function splitSlash(str) {
    const results = str.split('/')
        .map(s => s.trim().replace(/-/g, '_'));
    const maxResults = results.length < 2 ? results.length : 2;
    return results.slice(0, maxResults);
}
// get the flag variable names from the argparse commands
function getNames(flags) {
    return flags.map(flag => {
        return `_flag_${flag}`;
    });
}
/**
 * Process an argparse command and return all of the flag definitions as a `FishSymbol[]`
 * @param document The LspDocument we are processing
 * @param node The node we are processing, should be isCommandWithName(node, 'argparse')
 * @param children The children symbols of the current FishSymbol's we are processing (likely empty)
 * @returns An array of FishSymbol's that represent the flags defined in the argparse command
 */
function processArgparseCommand(document, node, children = []) {
    const result = [];
    // get the scope modifier
    const modifier = getArgparseScopeModifier(document, node);
    // array of nodes that are `argparse` flags
    const focuesedNodes = findArgparseDefinitionNames(node);
    // build the flags, and store them in the result array
    for (const n of focuesedNodes) {
        let flagNames = n.text;
        if (!flagNames)
            continue;
        // fixup the flag names for strings and concatenated flags
        if ((0, node_types_1.isString)(n))
            flagNames = flagNames.slice(1, -1);
        if (flagNames.includes('='))
            flagNames = flagNames.slice(0, flagNames.indexOf('='));
        // split the text into corresponding flags and convert them to `_flag_` format
        const seenFlags = splitSlash(flagNames);
        const names = getNames(seenFlags);
        // add all seenFlags to the `result: FishSymbol[]` array
        const flags = names.map((flagName, idx) => {
            const selectedRange = createSelectionRange(n, seenFlags, flagName, idx);
            return symbol_1.FishSymbol.fromObject({
                name: flagName,
                node: node,
                focusedNode: n,
                fishKind: 'ARGPARSE',
                uri: document.uri,
                detail: n.text,
                range: (0, tree_sitter_1.getRange)(n),
                selectionRange: selectedRange,
                scope: definition_scope_1.DefinitionScope.create(node.parent, modifier),
                children,
            }).addAliasedNames(...names);
        });
        result.push(...flags);
    }
    return result;
}
