"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SymbolParser = exports.OptionsParser = exports.CompleteParser = exports.FunctionParser = exports.AliasParser = exports.ArgparseParser = exports.ForParser = exports.ReadParser = exports.SetParser = exports.parsers = exports.DefinitionNodeNames = exports.ParsingDefinitionNames = exports.NodeTypes = exports.VariableDefinitionKeywords = exports.Parsers = void 0;
exports.isVariableDefinitionName = isVariableDefinitionName;
exports.isFunctionDefinitionName = isFunctionDefinitionName;
exports.isAliasDefinitionName = isAliasDefinitionName;
exports.isExportVariableDefinitionName = isExportVariableDefinitionName;
exports.isDefinitionName = isDefinitionName;
const SetParser = __importStar(require("./set"));
exports.SetParser = SetParser;
const ReadParser = __importStar(require("./read"));
exports.ReadParser = ReadParser;
const ForParser = __importStar(require("./for"));
exports.ForParser = ForParser;
const ArgparseParser = __importStar(require("./argparse"));
exports.ArgparseParser = ArgparseParser;
const AliasParser = __importStar(require("./alias"));
exports.AliasParser = AliasParser;
const ExportParser = __importStar(require("./export"));
const FunctionParser = __importStar(require("./function"));
exports.FunctionParser = FunctionParser;
const CompleteParser = __importStar(require("./complete"));
exports.CompleteParser = CompleteParser;
const OptionsParser = __importStar(require("./options"));
exports.OptionsParser = OptionsParser;
const SymbolParser = __importStar(require("./symbol"));
exports.SymbolParser = SymbolParser;
exports.Parsers = {
    set: SetParser,
    read: ReadParser,
    for: ForParser,
    argparse: ArgparseParser,
    function: FunctionParser,
    complete: CompleteParser,
    options: OptionsParser,
    symbol: SymbolParser,
    export: ExportParser,
};
exports.VariableDefinitionKeywords = [
    'set',
    'read',
    'argparse',
    'for',
    'function',
    'export',
];
/**
 * Checks if a node is a variable definition name.
 * Examples of variable names include:
 * - `set -g -x foo '...'`      -> foo
 * - `read -l bar baz`          -> bar baz
 * - `argparse h/help -- $argv` -> h/help
 * - `for foo in _ `            -> i
 * - `export foo=bar`           -> foo
 */
function isVariableDefinitionName(node) {
    return SetParser.isSetVariableDefinitionName(node) ||
        ReadParser.isReadVariableDefinitionName(node) ||
        ArgparseParser.isArgparseVariableDefinitionName(node) ||
        ForParser.isForVariableDefinitionName(node) ||
        FunctionParser.isFunctionVariableDefinitionName(node) ||
        ExportParser.isExportVariableDefinitionName(node);
}
/**
 * Checks if a node is a function definition name.
 * Examples of function names include:
 * - `function baz; end;`       -> baz
 */
function isFunctionDefinitionName(node) {
    return FunctionParser.isFunctionDefinitionName(node);
}
/**
 * Checks if a node is a alias definition name.
 * - `alias foo '__foo'`        -> foo
 * - `alias bar='__bar'`        -> bar
 */
function isAliasDefinitionName(node) {
    return AliasParser.isAliasDefinitionName(node);
}
/**
 * Checks if a node is a export definition name.
 * - `export foo=__foo`          -> foo
 * - `export bar='__bar'`        -> bar
 */
function isExportVariableDefinitionName(node) {
    return ExportParser.isExportVariableDefinitionName(node);
}
/**
 * Checks if a node is a definition name.
 * Definition names are variable names (read/set/argparse/function flags), function names (alias/function),
 */
function isDefinitionName(node) {
    return isVariableDefinitionName(node) || isFunctionDefinitionName(node) || isAliasDefinitionName(node);
}
exports.NodeTypes = {
    isVariableDefinitionName: isVariableDefinitionName,
    isFunctionDefinitionName: isFunctionDefinitionName,
    isAliasDefinitionName: isAliasDefinitionName,
    isDefinitionName: isDefinitionName,
    isSetVariableDefinitionName: SetParser.isSetVariableDefinitionName,
    isReadVariableDefinitionName: ReadParser.isReadVariableDefinitionName,
    isForVariableDefinitionName: ForParser.isForVariableDefinitionName,
    isExportVariableDefinitionName: ExportParser.isExportVariableDefinitionName,
    isArgparseVariableDefinitionName: ArgparseParser.isArgparseVariableDefinitionName,
    isFunctionVariableDefinitionName: FunctionParser.isFunctionVariableDefinitionName,
    isMatchingOption: OptionsParser.isMatchingOption,
};
exports.ParsingDefinitionNames = {
    isSetVariableDefinitionName: SetParser.isSetVariableDefinitionName,
    isReadVariableDefinitionName: ReadParser.isReadVariableDefinitionName,
    isForVariableDefinitionName: ForParser.isForVariableDefinitionName,
    isArgparseVariableDefinitionName: ArgparseParser.isArgparseVariableDefinitionName,
    isFunctionVariableDefinitionName: FunctionParser.isFunctionVariableDefinitionName,
    isFunctionDefinitionName: FunctionParser.isFunctionDefinitionName,
    isAliasDefinitionName: AliasParser.isAliasDefinitionName,
    isExportDefinitionName: ExportParser.isExportVariableDefinitionName,
};
exports.DefinitionNodeNames = {
    isDefinitionName: isDefinitionName,
    isVariableDefinitionName: isVariableDefinitionName,
    isFunctionDefinitionName: isFunctionDefinitionName,
    isAliasDefinitionName: isAliasDefinitionName,
};
__exportStar(require("./options"), exports);
exports.parsers = Object.keys(exports.Parsers).map(key => exports.Parsers[key]);
