"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SetModifiers = exports.SetOptions = void 0;
exports.isSetDefinition = isSetDefinition;
exports.isSetVariableDefinitionName = isSetVariableDefinitionName;
exports.findSetChildren = findSetChildren;
exports.setModifierDetailDescriptor = setModifierDetailDescriptor;
exports.processSetCommand = processSetCommand;
const node_types_1 = require("../utils/node-types");
const options_1 = require("./options");
const symbol_1 = require("./symbol");
const definition_scope_1 = require("../utils/definition-scope");
exports.SetOptions = [
    options_1.Option.create('-U', '--universal'),
    options_1.Option.create('-g', '--global'),
    options_1.Option.create('-f', '--function'),
    options_1.Option.create('-l', '--local'),
    options_1.Option.create('-x', '--export'),
    options_1.Option.create('-u', '--unexport'),
    options_1.Option.long('--path'),
    options_1.Option.long('--unpath'),
    options_1.Option.create('-a', '--append'),
    options_1.Option.create('-p', '--prepend'),
    options_1.Option.create('-e', '--erase'),
    options_1.Option.create('-q', '--query'),
    options_1.Option.create('-n', '--names'),
    options_1.Option.create('-S', '--show'),
    options_1.Option.long('--no-event'),
    options_1.Option.create('-L', '--long'),
    options_1.Option.create('-h', '--help'),
];
// const setModifiers = SetOptions.filter(option => option.equalsRawLongOption('--universal', '--global', '--function', '--local'));
exports.SetModifiers = [
    options_1.Option.create('-U', '--universal'),
    options_1.Option.create('-g', '--global'),
    options_1.Option.create('-f', '--function'),
    options_1.Option.create('-l', '--local'),
];
function isSetDefinition(node) {
    return (0, node_types_1.isCommandWithName)(node, 'set') && !node.children.some(child => (0, options_1.isMatchingOption)(child, options_1.Option.create('-q', '--query')));
}
/**
 * checks if a node is the variable name of a set command
 * set -g -x foo '...'
 *           ^-- cursor is here
 */
function isSetVariableDefinitionName(node) {
    if (!node.parent || !isSetDefinition(node.parent))
        return false;
    const searchNodes = findSetChildren(node.parent);
    const definitionNode = searchNodes.find(n => !(0, node_types_1.isOption)(n));
    return !!definitionNode && definitionNode.equals(node);
}
function getFallbackModifierScope(document, node) {
    const autoloadType = document.getAutoloadType();
    switch (autoloadType) {
        case 'conf.d':
        case 'config':
            return (0, node_types_1.isTopLevelDefinition)(node) ? 'global' : 'local';
        case 'functions':
        default:
            return 'local';
    }
}
function findSetChildren(node) {
    const children = node.childrenForFieldName('argument');
    const firstNonOption = children.findIndex(child => !(0, node_types_1.isOption)(child));
    return children.slice(0, firstNonOption + 1);
}
function setModifierDetailDescriptor(nodee) {
    const options = (0, options_1.findOptions)(nodee.childrenForFieldName('argument'), exports.SetModifiers);
    const exportedOption = options.found.find(o => o.option.equalsRawOption('-x', '--export') || o.option.equalsRawOption('-u', '--unexport'));
    const exportedStr = exportedOption ? exportedOption.option.isOption('-x', '--export') ? 'exported' : 'unexported' : '';
    const modifier = options.found.find(o => o.option.equalsRawOption('-U', '-g', '-f', '-l'));
    if (modifier) {
        switch (true) {
            case modifier.option.isOption('-U', '--universal'):
                return ['universally scoped', exportedStr].filter(Boolean).join('; ');
            case modifier.option.isOption('-g', '--global'):
                return ['globally scoped', exportedStr].filter(Boolean).join('; ');
            case modifier.option.isOption('-f', '--function'):
                return ['function scoped', exportedStr].filter(Boolean).join('; ');
            case modifier.option.isOption('-l', '--local'):
                return ['locally scoped', exportedStr].filter(Boolean).join('; ');
            default:
                return ['', exportedStr].filter(Boolean).join('; ');
        }
    }
    return ['', exportedStr].filter(Boolean).join('; ');
}
function processSetCommand(document, node, children = []) {
    /** skip `set -q/--query` */
    if (!isSetDefinition(node))
        return [];
    // create the searchNodes, which are the nodes after the command name, but before the variable name
    const searchNodes = findSetChildren(node);
    // find the definition node, which should be the last node of the searchNodes
    const definitionNode = searchNodes.find(n => !(0, node_types_1.isOption)(n));
    const skipText = ['-', '$', '('];
    if (!definitionNode || skipText.some(t => definitionNode.text.startsWith(t)))
        return [];
    const modifierOption = (0, options_1.findOptionsSet)(node.childrenForFieldName('argument'), exports.SetModifiers).pop();
    let modifier = 'local';
    if (modifierOption) {
        modifier = (0, symbol_1.SetModifierToScopeTag)(modifierOption.option);
    }
    else {
        modifier = getFallbackModifierScope(document, node);
    }
    return [
        symbol_1.FishSymbol.create(definitionNode.text.toString(), node, definitionNode, 'SET', document.uri, node.text.toString(), definition_scope_1.DefinitionScope.create(node.parent, modifier), children),
    ];
}
