"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.findMatchingOptions = exports.isMatchingOption = exports.isExportVariableDefinitionName = exports.isAliasDefinitionName = exports.isFunctionDefinitionName = exports.isVariableDefinitionName = void 0;
exports.isVariableDefinition = isVariableDefinition;
exports.isComment = isComment;
exports.isShebang = isShebang;
exports.isFunctionDefinition = isFunctionDefinition;
exports.isCommand = isCommand;
exports.isTopLevelFunctionDefinition = isTopLevelFunctionDefinition;
exports.isTopLevelDefinition = isTopLevelDefinition;
exports.isDefinition = isDefinition;
exports.isCommandName = isCommandName;
exports.isProgram = isProgram;
exports.isError = isError;
exports.isForLoop = isForLoop;
exports.isIfStatement = isIfStatement;
exports.isElseStatement = isElseStatement;
exports.isConditional = isConditional;
exports.isIfOrElseIfConditional = isIfOrElseIfConditional;
exports.isPossibleUnreachableStatement = isPossibleUnreachableStatement;
exports.isClause = isClause;
exports.isStatement = isStatement;
exports.isBlock = isBlock;
exports.isEnd = isEnd;
exports.isScope = isScope;
exports.isSemicolon = isSemicolon;
exports.isNewline = isNewline;
exports.isBlockBreak = isBlockBreak;
exports.isString = isString;
exports.isStringCharacter = isStringCharacter;
exports.isEndStdinCharacter = isEndStdinCharacter;
exports.isEscapeSequence = isEscapeSequence;
exports.isLongOption = isLongOption;
exports.isShortOption = isShortOption;
exports.isOption = isOption;
exports.isJoinedShortOption = isJoinedShortOption;
exports.hasShortOptionCharacter = hasShortOptionCharacter;
exports.isPipe = isPipe;
exports.gatherSiblingsTillEol = gatherSiblingsTillEol;
exports.isBeforeCommand = isBeforeCommand;
exports.isVariableExpansion = isVariableExpansion;
exports.isVariableExpansionWithName = isVariableExpansionWithName;
exports.isVariable = isVariable;
exports.isCompleteFlagCommandName = isCompleteFlagCommandName;
exports.findPreviousSibling = findPreviousSibling;
exports.findParentCommand = findParentCommand;
exports.isConcatenation = isConcatenation;
exports.isAliasWithName = isAliasWithName;
exports.findParentFunction = findParentFunction;
exports.findParentVariableDefinitionKeyword = findParentVariableDefinitionKeyword;
exports.findForLoopVariable = findForLoopVariable;
exports.findSetDefinedVariable = findSetDefinedVariable;
exports.hasParent = hasParent;
exports.findParent = findParent;
exports.hasParentFunction = hasParentFunction;
exports.findFunctionScope = findFunctionScope;
exports.scopeCheck = scopeCheck;
exports.wordNodeIsCommand = wordNodeIsCommand;
exports.isSwitchStatement = isSwitchStatement;
exports.isCaseClause = isCaseClause;
exports.isReturn = isReturn;
exports.isConditionalCommand = isConditionalCommand;
exports.isCommandFlag = isCommandFlag;
exports.isRegexArgument = isRegexArgument;
exports.isUnmatchedStringCharacter = isUnmatchedStringCharacter;
exports.isPartialForLoop = isPartialForLoop;
exports.isInlineComment = isInlineComment;
exports.isCommandWithName = isCommandWithName;
exports.isReturnStatusNumber = isReturnStatusNumber;
exports.isCompleteCommandName = isCompleteCommandName;
const tree_sitter_1 = require("./tree-sitter");
const barrel_1 = require("../parsing/barrel");
const options_1 = require("../parsing/options");
const barrel_2 = require("../parsing/barrel");
Object.defineProperty(exports, "isVariableDefinitionName", { enumerable: true, get: function () { return barrel_2.isVariableDefinitionName; } });
Object.defineProperty(exports, "isFunctionDefinitionName", { enumerable: true, get: function () { return barrel_2.isFunctionDefinitionName; } });
Object.defineProperty(exports, "isAliasDefinitionName", { enumerable: true, get: function () { return barrel_2.isAliasDefinitionName; } });
Object.defineProperty(exports, "isExportVariableDefinitionName", { enumerable: true, get: function () { return barrel_2.isExportVariableDefinitionName; } });
/**
 * checks if a node is a variable definition. Current syntax tree from tree-sitter-fish will
 * only tokenize variable names if they are defined in a for loop. Otherwise, they are tokenized
 * with the node type of 'name'.
 *
 * @param {SyntaxNode} node - the node to check if it is a variable definition
 * @returns {boolean} true if the node is a variable definition, false otherwise
 */
function isVariableDefinition(node) {
    return (0, barrel_2.isVariableDefinitionName)(node);
}
/**
 * fish shell comment: '# ...'
 */
function isComment(node) {
    return node.type === 'comment' && !isShebang(node);
}
function isShebang(node) {
    const parent = node.parent;
    if (!parent || !isProgram(parent)) {
        return false;
    }
    const firstLine = parent.firstChild;
    if (!firstLine) {
        return false;
    }
    if (!node.equals(firstLine)) {
        return false;
    }
    return (firstLine.type === 'comment' &&
        firstLine.text.startsWith('#!') &&
        firstLine.text.includes('fish'));
}
/**
 * function some_fish_func
 *     ...
 * end
 * @see isFunctionDefinitionName()
 */
function isFunctionDefinition(node) {
    return node.type === 'function_definition';
}
/**
 * checks for all fish types of SyntaxNodes that are commands.
 */
function isCommand(node) {
    return [
        'command',
        'test_command',
        'command_substitution',
    ].includes(node.type);
}
function isTopLevelFunctionDefinition(node) {
    if (isFunctionDefinition(node)) {
        return node.parent?.type === 'program';
    }
    if ((0, barrel_2.isFunctionDefinitionName)(node)) {
        return node.parent?.parent?.type === 'program';
    }
    return false;
}
function isTopLevelDefinition(node) {
    let currentNode = node;
    while (currentNode) {
        if (!currentNode)
            break;
        if (isProgram(currentNode)) {
            return true;
        }
        if (isFunctionDefinition(currentNode)) {
            return false;
        }
        currentNode = currentNode.parent;
    }
    return true;
}
/**
 * isVariableDefinitionName() || isFunctionDefinitionName()
 */
function isDefinition(node) {
    return (0, barrel_2.isFunctionDefinitionName)(node) || (0, barrel_2.isVariableDefinitionName)(node);
}
/**
 * checks if a node is the firstNamedChild of a command
 */
function isCommandName(node) {
    const parent = node.parent || node;
    const cmdName = parent?.firstNamedChild || node?.firstNamedChild;
    if (!parent || !cmdName) {
        return false;
    }
    if (!isCommand(parent)) {
        return false;
    }
    return node.type === 'word' && node.equals(cmdName);
}
/**
 * the root node of a fish script
 */
function isProgram(node) {
    return node.type === 'program' || node.parent === null;
}
function isError(node = null) {
    if (node) {
        return node.type === 'ERROR';
    }
    return false;
}
function isForLoop(node) {
    return node.type === 'for_statement';
}
function isIfStatement(node) {
    return node.type === 'if_statement';
}
function isElseStatement(node) {
    return node.type === 'else_clause';
}
// strict check for if statement or else clauses
function isConditional(node) {
    return ['if_statement', 'else_if_clause', 'else_clause'].includes(node.type);
}
function isIfOrElseIfConditional(node) {
    return ['if_statement', 'else_if_clause'].includes(node.type);
}
function isPossibleUnreachableStatement(node) {
    if (isIfStatement(node)) {
        return node.lastNamedChild?.type === 'else_clause';
    }
    else if (node.type === 'for_statement') {
        return true;
    }
    else if (node.type === 'switch_statement') {
        return false;
    }
    return false;
}
function isClause(node) {
    return [
        'case_clause',
        'else_clause',
        'else_if_clause',
    ].includes(node.type);
}
/**
 * statements contain clauses
 */
function isStatement(node) {
    return [
        'for_statement',
        'switch_statement',
        'while_statement',
        'if_statement',
        'begin_statement',
    ].includes(node.type);
}
/**
 * since statement SyntaxNodes contains clauses, treats statements and clauses the same:
 * if ...           - if_statement
 * else if ...      --- else_if_clause
 * else ...         --- else_clause
 * end;
 */
function isBlock(node) {
    return isClause(node) || isStatement(node);
}
function isEnd(node) {
    return node.type === 'end';
}
//export function isLocalBlock(node: SyntaxNode): boolean {
//return ['begin_statement'].includes(node.type);
//}
/**
 * Any SyntaxNode that will enclose a new local scope:
 *      Program, Function, if, for, while
 */
function isScope(node) {
    return isProgram(node) || isFunctionDefinition(node) || isStatement(node); // || isLocalBlock(node)//
}
function isSemicolon(node) {
    return node.type === ';' && node.text === ';';
}
function isNewline(node) {
    return node.type === '\n';
}
function isBlockBreak(node) {
    return isEnd(node) || isSemicolon(node) || isNewline(node);
}
function isString(node) {
    return [
        'double_quote_string',
        'single_quote_string',
    ].includes(node.type);
}
function isStringCharacter(node) {
    return [
        "'",
        '"',
    ].includes(node.type);
}
/**
 * Checks if a node is fish's end stdin token `--`
 * This is used to signal the end of stdin input, like in the argparse command: `argparse h/help -- $argv`
 * @param {SyntaxNode} node - the node to check
 * @returns  true if the node is the end stdin token
 */
function isEndStdinCharacter(node) {
    return '--' === node.text && node.type === 'word';
}
/**
 * Checks if a node is fish escape sequence token `\` character
 * This token will be used to escape commands which span multiple lines
 */
function isEscapeSequence(node) {
    return node.type === 'escape_sequence';
}
function isLongOption(node) {
    return node.text.startsWith('--') && !isEndStdinCharacter(node);
}
/**
 * node.text !== '-' because `-` this would not be an option... Consider the case:
 * ```
 * cat some_file | nvim -
 * ```
 */
function isShortOption(node) {
    return node.text.startsWith('-') && !isLongOption(node) && node.text !== '-';
}
/**
 * Checks if a node is an option/switch/flag in any of the following formats:
 *    - short options: `-g`, `-f1`, `-f 1`, `-f=2`, `-gx`
 *    - long options: `--global`, `--file`, `--file=1`, `--file 1`
 *    - old unix style flags: `-type`, `-type=file`
 * @param {SyntaxNode} node - the node to check
 * @returns {boolean} true if the node is an option
 */
function isOption(node) {
    if (isEndStdinCharacter(node))
        return false;
    return isShortOption(node) || isLongOption(node);
}
/** careful not to call this on old unix style flags/options */
function isJoinedShortOption(node) {
    if (isLongOption(node))
        return false;
    return isShortOption(node) && node.text.slice(1).length > 1;
}
/** careful not to call this on old unix style flags/options */
function hasShortOptionCharacter(node, findChar) {
    if (isLongOption(node))
        return false;
    return isShortOption(node) && node.text.slice(1).includes(findChar);
}
var options_2 = require("../parsing/options");
Object.defineProperty(exports, "isMatchingOption", { enumerable: true, get: function () { return options_2.isMatchingOption; } });
Object.defineProperty(exports, "findMatchingOptions", { enumerable: true, get: function () { return options_2.findMatchingOptions; } });
function isPipe(node) {
    return node.type === 'pipe';
}
function gatherSiblingsTillEol(node) {
    const siblings = [];
    let next = node.nextSibling;
    while (next && !isNewline(next)) {
        siblings.push(next);
        next = next.nextSibling;
    }
    return siblings;
}
/*
 * Checks for nodes which should stop the search for
 * command nodes, used in findParentCommand()
 */
function isBeforeCommand(node) {
    return [
        'file_redirect',
        'redirect',
        'redirected_statement',
        'conditional_execution',
        'stream_redirect',
        'pipe',
    ].includes(node.type) || isFunctionDefinition(node) || isStatement(node) || isSemicolon(node) || isNewline(node) || isEnd(node);
}
function isVariableExpansion(node) {
    return node.type === 'variable_expansion';
}
/**
 * Checks for variable expansions that match the variable name, DONT PASS `variableName` with leading `$`
 * @param {SyntaxNode} node - the node to check
 * @param {string} variableName - the name of the variable to check for (`pipestatus`, `status`, `argv`, ...)
 * @returns {boolean} true if the node is a variable expansion matching the name
 */
function isVariableExpansionWithName(node, variableName) {
    return node.type === 'variable_expansion' && node.text === `$${variableName}`;
}
function isVariable(node) {
    if (isVariableDefinition(node)) {
        return true;
    }
    else {
        return ['variable_expansion', 'variable_name'].includes(node.type);
    }
}
function isCompleteFlagCommandName(node) {
    if (node.parent && isCommandWithName(node, 'set')) {
        const children = node.parent.childrenForFieldName('arguments').filter(n => !isOption(n));
        if (children && children.at(0)?.equals(node)) {
            return node.text.startsWith('_flag_');
        }
    }
    return false;
}
/**
 * finds the parent command of the current node
 *
 * @param {SyntaxNode} node - the node to check for its parent
 * @returns {SyntaxNode | null} command node or null
 */
function findPreviousSibling(node) {
    let currentNode = node;
    if (!currentNode) {
        return null;
    }
    while (currentNode !== null) {
        if (isCommand(currentNode)) {
            return currentNode;
        }
        currentNode = currentNode.parent;
    }
    return null;
}
/**
 * finds the parent command of the current node
 *
 * @param {SyntaxNode} node - the node to check for its parent
 * @returns {SyntaxNode | null} command node or null
 */
function findParentCommand(node) {
    let currentNode = node;
    if (!currentNode) {
        return null;
    }
    while (currentNode !== null) {
        if (isCommand(currentNode)) {
            return currentNode;
        }
        currentNode = currentNode.parent;
    }
    return null;
}
function isConcatenation(node) {
    return node.type === 'concatenation';
}
function isAliasWithName(node, aliasName) {
    if ((0, barrel_2.isAliasDefinitionName)(node)) {
        return node.text.split('=').at(0) === aliasName;
    }
    return false;
}
/**
 * finds the parent function of the current node
 *
 * @param {SyntaxNode} node - the node to check for its parent
 * @returns {SyntaxNode | null} command node or null
 */
function findParentFunction(node) {
    let currentNode = node;
    if (!currentNode) {
        return null;
    }
    while (currentNode !== null) {
        if (isFunctionDefinition(currentNode)) {
            return currentNode;
        }
        currentNode = currentNode.parent;
    }
    return null;
}
function findParentVariableDefinitionKeyword(node) {
    if (!node || !(0, barrel_2.isVariableDefinitionName)(node))
        return null;
    const currentNode = node;
    const parent = currentNode?.parent;
    if (!currentNode || !parent) {
        return null;
    }
    const varKeyword = parent.firstChild?.text.trim() || '';
    if (!varKeyword) {
        return null;
    }
    if (barrel_1.VariableDefinitionKeywords.includes(varKeyword)) {
        return parent;
    }
    return null;
}
function findForLoopVariable(node) {
    for (let i = 0; i < node.children.length; i++) {
        const child = node.children[i];
        if (child?.type === 'variable_name') {
            return child;
        }
    }
    return null;
}
/**
 * @param {SyntaxNode} node - finds the node in a fish command that will
 *                            contain the variable definition
 *
 * @return {SyntaxNode | null} variable node that was found
 **/
function findSetDefinedVariable(node) {
    const parent = findParentCommand(node);
    if (!parent) {
        return null;
    }
    const children = parent.children;
    let i = 1;
    let child = children[i];
    while (child !== undefined) {
        if (!child.text.startsWith('-')) {
            return child;
        }
        if (i === children.length - 1) {
            return null;
        }
        child = children[i++];
    }
    return child;
}
function hasParent(node, callbackfn) {
    let currentNode = node;
    while (currentNode !== null) {
        if (callbackfn(currentNode)) {
            return true;
        }
        currentNode = currentNode.parent;
    }
    return false;
}
function findParent(node, callbackfn) {
    let currentNode = node;
    while (currentNode !== null) {
        if (callbackfn(currentNode)) {
            return currentNode;
        }
        currentNode = currentNode.parent;
    }
    return null;
}
function hasParentFunction(node) {
    let currentNode = node;
    while (currentNode !== null) {
        if (isFunctionDefinition(currentNode) || currentNode.type === 'function') {
            return true;
        }
        if (currentNode.parent === null) {
            return false;
        }
        currentNode = currentNode?.parent;
    }
    return false;
}
function findFunctionScope(node) {
    while (node.parent !== null) {
        if (isFunctionDefinition(node)) {
            return node;
        }
        node = node.parent;
    }
    return node;
}
// node1 encloses node2
function scopeCheck(node1, node2) {
    const scope1 = findFunctionScope(node1);
    const scope2 = findFunctionScope(node2);
    if (isProgram(scope1)) {
        return true;
    }
    return scope1 === scope2;
}
function wordNodeIsCommand(node) {
    if (node.type !== 'word') {
        return false;
    }
    return node.parent ? isCommand(node.parent) && node.parent.firstChild?.text === node.text : false;
}
function isSwitchStatement(node) {
    return node.type === 'switch_statement';
}
function isCaseClause(node) {
    return node.type === 'case_clause';
}
function isReturn(node) {
    return node.type === 'return' && node.firstChild?.text === 'return';
    //return node.type === 'return'
}
function isConditionalCommand(node) {
    return node.type === 'conditional_execution';
}
// @TODO: see ./tree-sitter.ts -> getRangeWithPrecedingComments(),
//        for implementation of chained returns of conditional_executions
// export function chainedCommandGroup(): SyntaxNode[] {
//   return [];
// }
function isCommandFlag(node) {
    return [
        'test_option',
        'word',
        'escape_sequence',
    ].includes(node.type) || node.text.startsWith('-') || findParentCommand(node) !== null;
}
function isRegexArgument(n) {
    return n.text === '--regex' || n.text === '-r';
}
function isUnmatchedStringCharacter(node) {
    if (!isStringCharacter(node)) {
        return false;
    }
    if (node.parent && isString(node.parent)) {
        return false;
    }
    return true;
}
function isPartialForLoop(node) {
    const semiCompleteForLoop = ['for', 'i', 'in', '_'];
    const errorNode = node.parent;
    if (node.text === 'for' && node.type === 'for') {
        if (!errorNode) {
            return true;
        }
        if ((0, tree_sitter_1.getLeafNodes)(errorNode).length < semiCompleteForLoop.length) {
            return true;
        }
        return false;
    }
    if (!errorNode) {
        return false;
    }
    return (errorNode.hasError &&
        errorNode.text.startsWith('for') &&
        !errorNode.text.includes(' in '));
}
function isInlineComment(node) {
    if (!isComment(node))
        return false;
    const previousSibling = node.previousNamedSibling;
    if (!previousSibling)
        return false;
    return previousSibling?.startPosition.row === node.startPosition.row && previousSibling?.type !== 'comment';
}
function isCommandWithName(node, ...commandNames) {
    if (node.type !== 'command')
        return false;
    // const currentCommandName = node.firstChild?.text
    return !!node.firstChild && commandNames.includes(node.firstChild.text);
}
function isReturnStatusNumber(node) {
    if (node.type !== 'integer')
        return false;
    const parent = node.parent;
    if (!parent)
        return false;
    return parent.type === 'return';
}
function isCompleteCommandName(node) {
    if (!node.parent || !isCommand(node.parent))
        return false;
    if (!isCommandWithName(node.parent, 'complete'))
        return false;
    const previousSibling = node.previousNamedSibling;
    if (!previousSibling)
        return false;
    if ((0, options_1.isMatchingOption)(previousSibling, options_1.Option.create('-c', '--command').withValue())) {
        return !isOption(node);
    }
    return false;
}
